<?php

/**
 * @author "Ahmad Hejazee" <mngafa@gmail.com>
 * @file Zarinpal class handles all transactions related to zarinpal payment service.
 * 
 * @see http://www.hejazee.ir/
 * @see http://www.tolooco.com/
 *
 * Copyright 2014-2016 - www.tolooco.com - All rights reserved.
 * For Licence, see README.txt and README.fa.txt
 */

/**
 * Class Zarinpal Handles Zarinpal Gateway transactions
 */
class Zarinpal {

  /**
   * @var bool Whether we should use nusoap library
   */
  protected $nusoap;

  /**
   * @var string Merchand ID
   */
  protected $MerchantID;

  /**
   * @var int Amount in Tomans
   */
  protected $Amount;

  /**
   * @var string Description of the transaction
   */
  protected $Description;

  /**
   * @var string optional Email of the customer
   */
  protected $Email;

  /**
   * @var string optional Mobile number of the customer
   */
  protected $Mobile;

  /**
   * @var Callback URL of the shopping site
   */
  protected $CallbackURL;

  /**
   * @var string Zarinpal webgate WSDL address in Germany
   */
  protected $webgate_auth_de = 'https://de.zarinpal.com/pg/services/WebGate/wsdl';

  /**
   * @var string Zarinpal webgate WSDL address in Iran
   */
  protected $webgate_auth_ir = 'https://ir.zarinpal.com/pg/services/WebGate/wsdl';

  /**
   * @var string Zarinpal Webgate address prefix
   */
  protected $webgate = 'https://www.zarinpal.com/pg/StartPay/';

  /**
   * @var bool Whether we should use iran switch or not (germany otherwise)
   */
  protected $Iran = False;

  /**
   * @var array List of copartners for checkout (used for zarinpal's 'Extra' methods)
   *  It will be disabled if it's an empty array (default)
   */
  protected $coPartners = array();

  /**
   * @var bool IF set to TRUE, PaymentVerificationWithExtra() will be called for verification.
   */
  protected $enableVerifyWithExtra = FALSE;

  /**
   * Class Constructor
   *
   * @param $MerchantID
   *   Merchant ID
   * @param $Amount
   *   Amount In tomans
   * @param $Description
   *   Description of Payment
   * @param $CallbackURL
   *   Callback url of the Shopping site
   * @param string $Email
   *   Optional email of customer
   * @param string $Mobile
   *   Optional mobile number of the customer
   */
  function __construct($MerchantID, $Amount, $Description, $CallbackURL, $Email= '', $Mobile = '') {
    //Check if we should use SoapClient or Nusoap
    $soapmode = variable_get('uc_zarinpal_soapmode', 'nusoap');
    if ($soapmode == 'soapclient' && class_exists('SoapClient')) {
      $this->nusoap = FALSE;
    }
    else {
      //Load the nusoap class.
      require_once dirname(__FILE__) . '/lib/nusoap/nusoap.php';
      $this->nusoap = TRUE;
    }

    //Initialize variables
    $this->MerchantID = $MerchantID;
    $this->Amount = $Amount;
    $this->Description = $Description;
    $this->CallbackURL = $CallbackURL;
    $this->Email = $Email;
    $this->Mobile = $Mobile;
  }

  /**
   * Set the zarinpal switch to Iran or Germany or check it.
   *
   * @param $arg
   *   If True, the switch will be set to Iran
   *   If False, the switch will be set to Germany
   *   Otherwise, the switch status will be returned (True = Iran; False = Germany)
   * @return bool
   */
  public function isIran($arg = NULL) {
    if ($arg === TRUE) {
      $this->Iran = TRUE;
    }
    elseif ($arg === FALSE) {
      $this->Iran = FALSE;
    }
    else {
      return $this->Iran;
    }

    return TRUE;
  }

  /**
   * Add a new copartner for checkout.
   *
   * If you use this method, Payment requests and verifications will be done with zarinpal's "Extra" methods.
   *
   * @param int $zpAccountNo
   *   Zarinpal account number.
   * @param int $Wallet
   *   Zarinpal wallet number.
   * @param int $Amount
   *   The shared amount (In price unit)
   * @param string $Desc
   *   A description for this share.
   */
  public function addCoPartner($zpAccountNo, $Wallet, $Amount, $Desc) {
    $zp_account_full = 'zp.' . $zpAccountNo . '.' . $Wallet;

    $this->coPartners['Wages'][$zp_account_full] = array(
      'Amount' => $Amount,
      'Description' => $Desc,
    );

    $this->enableVerifyWithExtra = TRUE;
  }

  /**
   * Check if copartners are enabled.
   *
   * @param bool $set If TRUE, It will be set to have coPartners.
   *
   * @return bool
   */
  public function hasCoPartners($set = FALSE) {
    if ($set) {
      $this->enableVerifyWithExtra = TRUE;

      return TRUE;
    }

    return !empty($this->coPartners) || $this->enableVerifyWithExtra;
  }

  /**
   * Format coPartners for zarinpal's PaymentRequestWithExtra() method.
   *
   * @return string JSON
   *  If copartners are disabled, returns FALSE.
   */
  protected function formatAdditionalData() {
    if (!$this->hasCoPartners()) {
      return FALSE;
    }

    return json_encode($this->coPartners);
  }

  /**
   * Request payment.
   *
   * @throws ZarinpalException_PaymentRequest
   *
   * @return String Authority number
   */
  public function paymentRequest() {
    $wsdl = $this->isIran() ? $this->webgate_auth_ir : $this->webgate_auth_de;
    $authority = NULL;

    $args = array(
      'MerchantID'  => $this->MerchantID,
      'Amount'      => $this->Amount,
      'Description' => $this->Description,
      'Email'       => $this->Email,
      'Mobile'      => $this->Mobile,
      'CallbackURL' => $this->CallbackURL
    );

    // Check if coPartners exist.
    if ($this->hasCoPartners()) {
      $args['AdditionalData'] = $this->formatAdditionalData();
    }

    if ($this->nusoap) {
      $client = new nusoap_client($wsdl, 'wsdl');
      $client->soap_defencoding = 'UTF-8';

      // If coPartners exist, we should use Extra Methods.
      if ($this->hasCoPartners()) {
        $result = $client->call('PaymentRequestWithExtra', array($args));
      }
      else {
        $result = $client->call('PaymentRequest', array($args));
      }

      if (($result['Status'] == 100) && (isset($result['Authority'])) && (strlen($result['Authority']) == 36)) {
        $authority = $result['Authority'];
      }
      else {
        throw new ZarinpalException_PaymentRequest($result['Status'], $result, $args);
      }
    }
    else {
      $client = new SoapClient($wsdl, array('encoding' => 'UTF-8'));

      // If coPartners exist, we should use Extra Methods.
      if ($this->hasCoPartners()) {
        $result = $client->PaymentRequestWithExtra($args);
      }
      else {
        $result = $client->PaymentRequest($args);
      }

      if (($result->Status == 100) && (isset($result->Authority)) && (strlen($result->Authority) == 36)) {
        $authority = $result->Authority;
      }
      else {
        throw new ZarinpalException_PaymentRequest($result->Status, (array)$result, $args);
      }
    }

    return $authority;
  }

  /**
   * Verify transaction
   *
   * @param String $Authority
   *   The Authority number which is passed by Zarinpal to the callback page as QueryString
   *   If Null provided, it will be get from $_GET
   *
   * @throws ZarinpalException_Verification_AuthorityMissing
   * @throws ZarinpalException_Verification_Nok
   * @throws ZarinpalException_Verification_InvalidStatus
   * @throws ZarinpalException_Verification_NotVerified
   *
   * @return int Reference ID of transaction if it is successful.
   */
  public function verifyTransaction($Authority = NULL) {
    // Validate $Authority and $Status (passed by zarinpal via QueryString in callback url)
    if (is_null($Authority)) {
      $status = isset($_GET['Status']) ? $_GET['Status'] : NULL;

      if ($status == 'OK') {
        if (isset($_GET['Authority'])) {
          $Authority = $_GET['Authority'];
        }
        else {
          throw new ZarinpalException_Verification_AuthorityMissing();
        }
      }
      elseif ($status == 'NOK' or $status == 'CANCELED') {
        // Transaction canceled by user.
        throw new ZarinpalException_Verification_Nok();
      }
      else {
        throw new ZarinpalException_Verification_InvalidStatus($status);
      }
    }

    // Validated. Now Process verification
    $wsdl = $this->isIran() ? $this->webgate_auth_ir : $this->webgate_auth_de;
    $refid = NULL;
    $args = array(
      'MerchantID' => $this->MerchantID,
      'Amount'     => $this->Amount,
      'Authority'  => $Authority,
    );

    if ($this->nusoap) {
      $client = new nusoap_client($wsdl, 'wsdl');
      $client->soap_defencoding = 'UTF-8';

      // If coPartners exist, we should use Extra Methods.
      if ($this->hasCoPartners()) {
        $res = $client->call('PaymentVerificationWithExtra', array($args));
      }
      else {
        $res = $client->call('PaymentVerification', array($args));
      }

      if (isset($res['Status']) && ($res['Status'] == 100)) {
        $refid = $res['RefID'];
        $extra_detail = isset($res['ExtraDetail']) ? $res['ExtraDetail'] : NULL;
      }
      else {
        throw new ZarinpalException_Verification_NotVerified($res);
      }

    }
    else {
      $client = new SoapClient($wsdl, array('encoding' => 'UTF-8'));

      // If coPartners exist, we should use Extra Methods.
      if ($this->hasCoPartners()) {
        $res = $client->PaymentVerificationWithExtra($args);
      }
      else {
        $res = $client->PaymentVerification($args);
      }

      if (isset($res->Status) && ($res->Status == 100)) {
        $refid = $res->RefID;
        $extra_detail = isset($res->ExtraDetail) ? $res->ExtraDetail : NULL;
      }
      else {
        throw new ZarinpalException_Verification_NotVerified((array)$res);
      }
    }

    return array($refid, $extra_detail);
  }

  /**
   * Build WebGate or ZarinGate URL. The user should be redirected to this address for payment
   *
   * @param String $Authority The authority code returned by paymentRequest() method
   * @param string $type Type of Gate. Wither 'webgate' or 'zaringate'. default is 'webgate'
   *
   * @return string Gate url
   *
   * @throws Exception
   */
  public function buildGateURL($Authority, $type = 'webgate') {
    switch ($type) {
      case 'webgate':
        $url = $this->webgate . $Authority;
        break;
      case 'zaringate':
        $url = $this->webgate . $Authority . '/ZarinGate';
        break;
      default:
        throw new Exception('Invalid gate type: should be either "webgate" or "zaringate"');
    }

    return $url;
  }

  /**
   * Translate & localize Zarinpal error codes
   *
   * @param int $code The error code
   *
   * @return string Translated & localized message
   */
  public static function translateErrorCode($code) {
    switch ($code) {
      case -1:
        $msg = t('Incomplete data.');
        break;
      case -2:
        $msg = t('Invalid IP or MerchantID.');
        break;
      case -3:
        $msg = t('Cannot pay this amount due to restrictions of the Shapark system.');
        break;
      case -4:
        $msg = t('Merchant\'s verify-level is below the accepted silver level.');
        break;
      case -11:
        $msg = t('Specified request was not found.');
        break;
      case -12:
        $msg = t('Can\'t edit request.');
        break;
      case -21:
        $msg = t('No financial actions found for this transaction.');
        break;
      case -22:
        $msg = t('Transaction failed.');
        break;
      case -33:
        $msg = t('Transaction\'s amount does not match with the paid amount.');
        break;
      case -34:
        $msg = t('Maximum transaction division exceeded (by quantity or amount).');
        break;
      case -40:
        $msg = t('Access to requested method is denied.');
        break;
      case -41:
        $msg = t('Invalid data sent to AdditionalData() method.');
        break;
      case -42:
        $msg = t('Paymeny authority lifetime should be between 30 minutes and 45 days.');
        break;
      case -54:
        $msg = t('The specified request is archived.');
        break;
      case 100:
        $msg = t('Action completed successfully');
        break;
      case 101:
        $msg = t('Payment was successfull but PaymentVerification() method is already done.');
        break;
      default:
        $msg = t('Unknown error code.');
    }

    $message = t("Error code: @code : @msg", array(
      '@code' => $code,
      '@msg' => $msg,
    ));

    return $message;
  }
}

/**
 * Exception: When Payment Request fails.
 */
class ZarinpalException_PaymentRequest extends \Exception {
  /**
   * @var int The error code returned by Zarinpal's verification method.
   */
  private $error_code;

  /**
   * @var string Translation of the error code.
   */
  private $error_meaning;

  /**
   * @var array The result returned by Zarinpal's verification method.
   */
  private $res;

  /**
   * @var array Arguments send to the Zarinpal's verification method.
   */
  private $args;

  /**
   * @param int $error_code The error code returned by Zarinpal's verification method.
   * @param array $res The result returned by Zarinpal's verification method.
   * @param array $args Arguments send to the Zarinpal's verification method.
   */
  public function __construct($error_code, array $res, array $args) {
    $this->error_code = $error_code;
    $this->error_meaning = Zarinpal::translateErrorCode($error_code);
    $this->res = $res;
    $this->args = $args;

    // Build message.
    $message = t('Can not request payment. !msg. Full $res: !res, Args: !args', array(
      '!msg'  => $this->error_meaning,
      '!res'  => print_r($this->res, TRUE),
      '!args' => print_r($this->args, TRUE),
    ));

    parent::__construct($message);
  }

  /**
   * Get a message suitable for displaying to user.
   */
  public function getUserMessage() {
    $message = t('Payment request failed. Please try again later.');

    return $message;
  }
}

/**
 * Exception: When $_GET[Authority] is not available during verification.
 */
class ZarinpalException_Verification_AuthorityMissing extends \Exception{
  public function __construct() {
    parent::__construct(t('Payment Verification error: $_GET[Authority] not available'));
  }

  public function getUserMessage() {
    return t('Zarinpal failed to send authority. Please contact administrator.');
  }
}

/**
 * Custom Exception class
 * When Zarinpal sends Status=NOK to callback URL as query string.
 * In Sample PHP Code: Transaction canceled by user.
 */
class ZarinpalException_Verification_Nok extends Exception {
  public function __construct() {
    parent::__construct(t('Payment Verification error: $_GET[Status] is NOK or CANCELED.'));
  }

  public function getUserMessage() {
    return t('Transaction not successful or cancelled by user.');
  }
}

/**
 * Exception: When $_GET[Status] is invalid during verification
 */
class ZarinpalException_Verification_InvalidStatus extends \Exception {
  private $status;

  /**
   * @param string $status The invalid status
   */
  public function __construct($status) {
    $this->status = $status;
    $message = t('Error: $_GET[Status] invalid: @status', array('@status' => $status));

    parent::__construct($message);
  }

  public function getUserMessage() {
    return t('Zarinpal response status is invalid. Please contact administrator.');
  }
}

/**
 * Exception: When Verification fails.
 *  ($res['Status'] is not set or is not 100)
 */
class ZarinpalException_Verification_NotVerified extends \Exception {

  /**
   * @var array $res The result array returned by Zarinpal's verification method
   */
  private $res;

  private $refid;

  private $status;

  private $status_meaning;

  public function __construct(array $result) {
    $this->res = $result;
    $this->refid = $result['RefID'];
    $this->status = isset($result['Status']) ? $result['Status'] : NULL;

    $this->status_meaning = '';
    if ($this->status) {
      $this->status_meaning = Zarinpal::translateErrorCode($result['Status']);
    }

    // Build message.
    $message = t('Transaction verification failed. RefID: !refid, Message: !message, Status: !status, Full $res: !res', array(
      '!refid' => $this->refid,
      '!message' => $this->status_meaning,
      '!status' => $this->status ? $this->status : t('No staus available.'),
      '!res' => print_r($this->res, TRUE),
    ));

    parent::__construct($message);
  }

  public function getUserMessage() {
    return t('Transaction verification failed.') . ' ' . $this->status_meaning;
  }
}
